/*

  sshmp-integer.c

  Author: Mika Kojo <mkojo@ssh.fi>

  Copyright (C) 1996-2001 SSH Communications Security Corp
  All rights reserved.

  Created: Tue Oct 28 14:54:44 1997 [mkojo]

  */

/* This library implements the computations in the ring of integers,
   that is in the set

     Z = {..., -2, -1, 0, 1, 2, ...}.

   Namely, the ring operations and some convenience routines.

   The more arithmetical routines (i.e. routines that also utilize
   local mod p features) are implemented elsewhere.  */

#include "sshincludes.h"
#include "sshmp.h"

/********** Routines for handling variable length integers *****/

/* Routines for allocating and expanding SshMPInteger's. */

SshMPInteger ssh_mprz_malloc(void)
{
  SshMPInteger op;
  op = ssh_xmalloc(sizeof(*op));

  /* Initialize the SshMPInteger. */
  op->m = 0;
  op->n = 0;
  op->sign = 0;
  op->v = NULL;
  return op;
}

void ssh_mprz_free(SshMPInteger op)
{
  ssh_xfree(op->v);
  ssh_xfree(op);
}

void ssh_mprz_realloc(SshMPInteger op, unsigned int new_size)
{
  if (new_size > op->m)
    {
      SshWord *nv;

      /* Allocate, copy and clear the rest. */
      nv = ssh_xmalloc((size_t)new_size * sizeof(SshWord));
      ssh_mpk_memcopy(nv, op->v, op->n);

      /* Free the old one. */
      ssh_xfree(op->v);

      /* Set the new one. */
      op->v = nv;
      op->m = new_size;
    }
}

/* Clear the upper (part which is not used) part of the
   integer. This allows us to sometimes use the integer's own
   data area for computations. */
void ssh_mprz_clear_extra(SshMPInteger op)
{
  unsigned int i;
  for (i = op->n; i < op->m; i++)
    op->v[i] = 0;
}

/****************** The integer interface. ******************/

/* Initialize the integer. */
void ssh_mprz_init(SshMPInteger op)
{
  op->m = 0;
  op->n = 0;
  op->sign = 0;
  op->v = NULL;
}

/* Clear the integer up, free the space occupied, but don't free the
   integer context. */
void ssh_mprz_clear(SshMPInteger op)
{
  ssh_xfree(op->v);
  op->n = 0;
  op->m = 0;
  op->sign = 0;
  op->v = NULL;
}

SshWord ssh_mprz_get_ui(SshMPIntegerConst op)
{
  if (op->n == 0)
    return 0;
  return op->v[0];
}

SshSignedWord ssh_mprz_get_si(SshMPIntegerConst op)
{
  SshSignedWord si;
  if (op->n == 0)
    return 0;
  /* Figure the bits that can be used. */
  si = (SshSignedWord)(op->v[0] & (SSH_WORD_MASK >> 1));
  if (SSH_MP_GET_SIGN(op))
    return -si;
  return si;
}
























































































void ssh_mprz_set(SshMPInteger ret, SshMPIntegerConst op)
{
  /* Check that pointers are not equal, in which case, anything more
     would be stupid. */
  if (ret == op)
    return;

  ssh_mprz_realloc(ret, op->n);

  /* Copy */
  ssh_mpk_memcopy(ret->v, op->v, op->n);
  ret->n = op->n;
  SSH_MP_COPY_SIGN(ret, op);
}

void ssh_mprz_set_ui(SshMPInteger op, SshWord n)
{
  if (n == 0)
    {
      op->n = 0;
      SSH_MP_NO_SIGN(op);
      return;
    }

  /* Check that we have enough space. */
  ssh_mprz_realloc(op, 1);

  /* Set the integer. */
  op->v[0] = (SshWord)n;
  op->n = 1;
  SSH_MP_NO_SIGN(op);
}

void ssh_mprz_set_si(SshMPInteger op, SshSignedWord n)
{
  if (n == 0)
    {
      op->n = 0;
      SSH_MP_NO_SIGN(op);
      return;
    }

  /* Check that we have enough space. */
  ssh_mprz_realloc(op, 1);

  if (n < 0)
    {
      SSH_MP_SET_SIGN(op);
      n = -n;
    }
  else
    SSH_MP_NO_SIGN(op);
  /* Set the integer. */
  op->v[0] = (SshWord)n;
  op->n = 1;
}

/* Init and assign. */
void ssh_mprz_init_set(SshMPInteger ret, SshMPIntegerConst op)
{
  ssh_mprz_init(ret);
  ssh_mprz_set(ret, op);
}

int ssh_mprz_init_set_str(SshMPInteger ret, const char *str, unsigned int base)
{
  ssh_mprz_init(ret);
  return ssh_mprz_set_str(ret, str, base);
}

void ssh_mprz_init_set_ui(SshMPInteger ret, SshWord u)
{
  ssh_mprz_init(ret);
  ssh_mprz_set_ui(ret, u);
}

void ssh_mprz_init_set_si(SshMPInteger ret, SshSignedWord s)
{
  ssh_mprz_init(ret);
  ssh_mprz_set_si(ret, s);
}

/* Negate an integer. This is very easy operation. */
void ssh_mprz_neg(SshMPInteger ret, SshMPIntegerConst op)
{
  ssh_mprz_set(ret, op);
  if (ret->n)
    SSH_MP_XOR_SIGN(ret);
}

/* Get the absolute of an integer, basically a distance in Z. */
void ssh_mprz_abs(SshMPInteger ret, SshMPIntegerConst op)
{
  ssh_mprz_set(ret, op);
  SSH_MP_NO_SIGN(ret);
}

/* The sign of an integer. We follow here the standard practice of
   naming the function. However, some call it just sign, but in
   number theory I have read about it as signum. */
int ssh_mprz_signum(SshMPIntegerConst op)
{
  if (SSH_MP_GET_SIGN(op))
    return -1;
  return 1;
}

/* These routines are written to be quick enough to be used in general.
   In some particular cases faster ways might be available. */

void ssh_mprz_mul_2exp(SshMPInteger ret, SshMPIntegerConst op,
                       unsigned int bits)
{
  unsigned int k, i;

  /* Check if no need to to anything. */
  if (op->n == 0)
    {
      ssh_mprz_set_ui(ret, 0);
      return;
    }

  if (bits == 0)
    {
      ssh_mprz_set(ret, op);
      return;
    }

  k = bits / SSH_WORD_BITS;
  bits %= SSH_WORD_BITS;

  /* Move from op to ret. */
  ssh_mprz_set(ret, op);

  /* Allocate new space. */
  ssh_mprz_realloc(ret, k + 1 + ret->n);

  /* Move words first. */
  if (k)
    {
      for (i = ret->n; i; i--)
        ret->v[i + k - 1] = ret->v[i - 1];
      for (i = 0; i < k; i++)
        ret->v[i] = 0;
    }

  /* Set the possible highest word to zero. */
  ret->v[k + ret->n] = 0;
  ssh_mpk_shift_up_bits(ret->v + k, ret->n + 1, bits);

  /* Compute the correct size. */
  ret->n = ret->n + k + 1;

  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;

  /* Remember the sign thing. */
  if (ret->n == 0)
    SSH_MP_NO_SIGN(ret);
}

void ssh_mprz_div_2exp(SshMPInteger ret, SshMPIntegerConst op,
                       unsigned int bits)
{
  unsigned int k, i;

  /* Check sizes. */
  if (op->n == 0)
    {
      ssh_mprz_set_ui(ret, 0);
      return;
    }

  if (bits == 0)
    {
      ssh_mprz_set(ret, op);
      return;
    }

  k = bits / SSH_WORD_BITS;
  bits %= SSH_WORD_BITS;

  if (k >= op->n)
    {
      ret->n = 0;
      return;
    }

  /* Move from op to ret. */
  ssh_mprz_set(ret, op);

  /* Move down. */
  if (k)
    for (i = 0; i < ret->n - k; i++)
      ret->v[i] = ret->v[i + k];

  ssh_mpk_shift_down_bits(ret->v, ret->n - k, bits);

  /* Compute new size. */
  ret->n = ret->n - k;
  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;

  if (ret->n == 0)
    SSH_MP_NO_SIGN(ret);
}

void ssh_mprz_mod_2exp(SshMPInteger ret, SshMPIntegerConst op,
                       unsigned int bits)
{
  unsigned int k;

  /* Check for trivial cases. */
  if (op->n == 0)
    {
      ssh_mprz_set_ui(ret, 0);
      return;
    }

  if (bits == 0)
    {
      ssh_mprz_set_ui(ret, 0);
      return;
    }

  k = bits / SSH_WORD_BITS;
  bits %= SSH_WORD_BITS;

  /* Now copy to the ret. This might not be the optimal way but easy.  */
  ssh_mprz_set(ret, op);

  /* Check yet one more trivial case. We might be done already. */
  if (k >= ret->n)
    return;

  /* Now we have to do the very hard part. */
  ret->v[k] = (ret->v[k] & (((SshWord)1 << bits) - 1));

  /* Check sizes. */
  ret->n = k + 1;
  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;

  if (ret->n == 0)
    SSH_MP_NO_SIGN(ret);
}

/* Comparison function which use directly the ssh_mpk_* functions. */
int ssh_mprz_cmp(SshMPIntegerConst op1, SshMPIntegerConst op2)
{
  int rv;
  /* Handle signs. */
  if (SSH_MP_GET_SIGN(op1) || SSH_MP_GET_SIGN(op2))
    {
      if (SSH_MP_GET_SIGN(op1) && !SSH_MP_GET_SIGN(op2))
        return -1;
      if (!SSH_MP_GET_SIGN(op1) && SSH_MP_GET_SIGN(op2))
        return 1;
    }
  rv = ssh_mpk_cmp(op1->v, op1->n, op2->v, op2->n);
  if (SSH_MP_GET_SIGN(op1) || SSH_MP_GET_SIGN(op2))
    rv = -rv;
  return rv;
}

int ssh_mprz_cmp_ui(SshMPIntegerConst op, SshWord u)
{
  if (SSH_MP_GET_SIGN(op))
    return -1;
  return ssh_mpk_cmp_ui(op->v, op->n, u);
}

int ssh_mprz_cmp_si(SshMPIntegerConst op, SshSignedWord s)
{
  int rv;
  SshWord sw = 0L;
  if (SSH_MP_GET_SIGN(op) || (s < 0))
    {
      if (SSH_MP_GET_SIGN(op) && (s >= 0))
        return -1;
      if (!SSH_MP_GET_SIGN(op) && (s < 0))
        return 1;
      /* Make s positive. */
      if (s < 0)
        sw = (SshWord)(-s);
    }
  else
    sw = (SshWord)s;
  rv = ssh_mpk_cmp_ui(op->v, op->n, sw);
  if (SSH_MP_GET_SIGN(op) && (s < 0))
    rv = -rv;
  return rv;
}

/* Addition routine which handles signs. */

void ssh_mprz_add(SshMPInteger ret, SshMPIntegerConst  op1,
                  SshMPIntegerConst  op2)
{
  SshWord c;

  if (op1->n == 0)
    {
      ssh_mprz_set(ret, op2);
      return;
    }

  if (op2->n == 0)
    {
      ssh_mprz_set(ret, op1);
      return;
    }

  /* Make op1 > op2 in absolute value. Also enlarge ret so that the
     result fits into it. */

  if (op1->n < op2->n)
    {
      SshMPIntegerConst  t;

      t   = op1;
      op1 = op2;
      op2 = t;
    }

  if (op1->n + 1 > ret->n)
    ssh_mprz_realloc(ret, op1->n + 1);

  /* Then figure out which case it really is. This idea of
     switching cames from my small floating point library that I
     wrote year ago. */

  switch ((SSH_MP_GET_SIGN(op1) << 1) + SSH_MP_GET_SIGN(op2))
    {
    case 0:
      c = ssh_mpk_add(ret->v, op1->v, op1->n, op2->v, op2->n);
      if (c)
        {
          ret->v[op1->n] = c;
          ret->n = op1->n + 1;
        }
      else
        ret->n = op1->n;
      SSH_MP_NO_SIGN(ret);
      break;
    case 1:
      if (ssh_mpk_cmp(op1->v, op1->n, op2->v, op2->n) >= 0)
        {
          ssh_mpk_sub(ret->v, op1->v, op1->n, op2->v, op2->n);
          SSH_MP_NO_SIGN(ret);
        }
      else
        {
          ssh_mpk_sub(ret->v, op2->v, op2->n, op1->v, op1->n);
          SSH_MP_SET_SIGN(ret);
        }
      ret->n = op1->n;
      break;
    case 2:
      if (ssh_mpk_cmp(op1->v, op1->n, op2->v, op2->n) >= 0)
        {
          ssh_mpk_sub(ret->v, op1->v, op1->n, op2->v, op2->n);
          SSH_MP_SET_SIGN(ret);
        }
      else
        {
          ssh_mpk_sub(ret->v, op2->v, op2->n, op1->v, op1->n);
          SSH_MP_NO_SIGN(ret);
        }
      ret->n = op1->n;
      break;
    case 3:
      c = ssh_mpk_add(ret->v, op1->v, op1->n, op2->v, op2->n);
      if (c)
        {
          ret->v[op1->n] = c;
          ret->n = op1->n + 1;
        }
      else
        ret->n = op1->n;
      SSH_MP_SET_SIGN(ret);
      break;
    }

  /* Following code should be place into either a macro or a function. */

  /* Correct the size. */
  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;

  if (ret->n == 0)
    SSH_MP_NO_SIGN(ret);
}

/* Subtraction routine which handles signs. */

void ssh_mprz_sub(SshMPInteger ret, SshMPIntegerConst op1,
                SshMPIntegerConst op2)
{
  SshWord c;
  unsigned int signs;

  if (op2->n == 0)
    {
      ssh_mprz_set(ret, op1);
      return;
    }

  if (op1->n == 0)
    {
      ssh_mprz_neg(ret, op2);
      return;
    }

  /* Make op1 > op2 in absolute value. Also enlarge ret so that the
     result fits in it. */

  if (op1->n < op2->n)
    {
      SshMPIntegerConst t;

      t = op1;
      op1 = op2;
      op2 = t;

      signs = ((SSH_MP_GET_SIGN(op1) ^ 0x1) << 1) + SSH_MP_GET_SIGN(op2);
    }
  else
    signs = (SSH_MP_GET_SIGN(op1) << 1) + (SSH_MP_GET_SIGN(op2) ^ 0x1);

  if (op1->n + 1 > ret->n)
    ssh_mprz_realloc(ret, op1->n + 1);

  /* Then figure out which case it really is. Note the difference between
     addition and subtraction. */

  switch (signs)
    {
    case 0:
      c = ssh_mpk_add(ret->v, op1->v, op1->n, op2->v, op2->n);
      if (c)
        {
          ret->v[op1->n] = c;
          ret->n = op1->n + 1;
        }
      else
        ret->n = op1->n;
      /* No sign for ret. */
      SSH_MP_NO_SIGN(ret);
      break;
    case 1:
      if (ssh_mpk_cmp(op1->v, op1->n, op2->v, op2->n) >= 0)
        {
          ssh_mpk_sub(ret->v, op1->v, op1->n, op2->v, op2->n);
          SSH_MP_NO_SIGN(ret);
        }
      else
        {
          ssh_mpk_sub(ret->v, op2->v, op2->n, op1->v, op1->n);
          SSH_MP_SET_SIGN(ret);
        }
      ret->n = op1->n;
      break;
    case 2:
      if (ssh_mpk_cmp(op1->v, op1->n, op2->v, op2->n) >= 0)
        {
          ssh_mpk_sub(ret->v, op1->v, op1->n, op2->v, op2->n);
          SSH_MP_SET_SIGN(ret);
        }
      else
        {
          ssh_mpk_sub(ret->v, op2->v, op2->n, op1->v, op1->n);
          SSH_MP_NO_SIGN(ret);
        }
      ret->n = op1->n;
      break;
    case 3:
      c = ssh_mpk_add(ret->v, op1->v, op1->n, op2->v, op2->n);
      if (c)
        {
          ret->v[op1->n] = c;
          ret->n = op1->n + 1;
        }
      else
        ret->n = op1->n;
      SSH_MP_SET_SIGN(ret);
      break;
    }

  /* Following code should be place into either a macro or a function. */

  /* Correct the size. */
  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;

  if (ret->n == 0)
    SSH_MP_NO_SIGN(ret);
}

/* Addition of a SshMPInteger and an SshWord. */
void ssh_mprz_add_ui(SshMPInteger ret, SshMPIntegerConst op, SshWord u)
{
  SshWord c;

  if (op->n == 0)
    {
      ssh_mprz_set_ui(ret, u);
      return;
    }

  ssh_mprz_realloc(ret, op->n + 1);

  switch (SSH_MP_GET_SIGN(op))
    {
    case 0:
      c = ssh_mpk_add(ret->v, op->v, op->n, &u, 1);
      if (c)
        {
          ret->v[op->n] = c;
          ret->n = op->n + 1;
        }
      else
        ret->n = op->n;
      SSH_MP_NO_SIGN(ret);
      break;
    case 1:
      if (ssh_mpk_cmp_ui(op->v, op->n, u) > 0)
        {
          ssh_mpk_sub(ret->v, op->v, op->n, &u, 1);
          SSH_MP_SET_SIGN(ret);
        }
      else
        {
          ssh_mpk_sub(ret->v, &u, 1, op->v, op->n);
          SSH_MP_NO_SIGN(ret);
        }
      ret->n = op->n;
      break;
    }

  /* Check size. */
  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;

  if (ret->n == 0)
    SSH_MP_NO_SIGN(ret);
}

/* Subtraction of an unsigned integer from a SshMPInteger. */
void ssh_mprz_sub_ui(SshMPInteger ret, SshMPIntegerConst op, SshWord u)
{
  SshWord c;

  if (op->n == 0)
    {
      ssh_mprz_set_ui(ret, u);
      return;
    }

  ssh_mprz_realloc(ret, op->n + 1);

  switch (SSH_MP_GET_SIGN(op))
    {
    case 0:
      if (ssh_mpk_cmp_ui(op->v, op->n, u) > 0)
        {
          ssh_mpk_sub(ret->v, op->v, op->n, &u, 1);
          SSH_MP_NO_SIGN(ret);
        }
      else
        {
          ssh_mpk_sub(ret->v, &u, 1, op->v, op->n);
          SSH_MP_SET_SIGN(ret);
        }
      ret->n = op->n;
      break;
    case 1:
      c = ssh_mpk_add(ret->v, op->v, op->n, &u, 1);
      if (c)
        {
          ret->v[op->n] = c;
          ret->n = op->n + 1;
        }
      else
        ret->n = op->n;
      SSH_MP_SET_SIGN(ret);
      break;
    }

  /* Check size. */
  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;

  if (ret->n == 0)
    SSH_MP_NO_SIGN(ret);
}

/* Multiplication routine. */
void ssh_mprz_mul(SshMPInteger ret, SshMPIntegerConst op1,
                  SshMPIntegerConst op2)
{
  SSH_MP_WORKSPACE_DEFINE;
  SshWord *temp;
  unsigned int temp_n;

  /* Check the inputs. */
  if (op1->n == 0 || op2->n == 0)
    {
      ssh_mprz_set_ui(ret, 0);
      return;
    }

  /* Allocate some temporary memory. */
  temp_n = op1->n + op2->n + 1;
  ssh_mprz_realloc(ret, temp_n);

  if (op1->v == ret->v || op2->v == ret->v)
    {
      SSH_MP_WORKSPACE_ALLOC(temp, temp_n);
    }
  else
    temp = ret->v;

  ssh_mpk_memzero(temp, temp_n);

  /* Do the multiplication. */
  ssh_mpk_mul_karatsuba(temp, temp_n, op1->v, op1->n, op2->v, op2->n,
                        NULL, 0);

  /* Check the exact length of the result. */

  while (temp_n && temp[temp_n - 1] == 0)
    temp_n--;

  /* Check the sign. */
  SSH_MP_XOR_SIGNS(ret, op1, op2);

  /* Finish by copying result to ret. */
  if (ret->v != temp)
    {
      ssh_mpk_memcopy(ret->v, temp, temp_n);
      SSH_MP_WORKSPACE_FREE(temp);
    }

  ret->n = temp_n;
}

/* Squaring routine. */

void ssh_mprz_square(SshMPInteger ret, SshMPIntegerConst op)
{
  SSH_MP_WORKSPACE_DEFINE;
  SshWord *temp;
  unsigned int temp_n;

  /* Check the inputs. */
  if (op->n == 0)
    {
      ssh_mprz_set_ui(ret, 0);
      return;
    }

  /* Allocate some temporary memory. */
  temp_n = op->n * 2 + 2;
  ssh_mprz_realloc(ret, temp_n);

  if (op->v == ret->v)
    {
      SSH_MP_WORKSPACE_ALLOC(temp, temp_n);
    }
  else
    temp = ret->v;

  ssh_mpk_memzero(temp, temp_n);

  /* Do the multiplication. */
  ssh_mpk_square_karatsuba(temp, temp_n, op->v, op->n,
                           NULL, 0);

  /* Check the exact length of the result. */

  while (temp_n && temp[temp_n - 1] == 0)
    temp_n--;

  /* Squaring, thus no sign! */
  SSH_MP_NO_SIGN(ret);

  /* Finish by copying result to ret. */
  if (ret->v != temp)
    {
      ssh_mpk_memcopy(ret->v, temp, temp_n);
      SSH_MP_WORKSPACE_FREE(temp);
    }

  ret->n = temp_n;
}

/* Division routine. */

void ssh_mprz_divrem(SshMPInteger q, SshMPInteger r,
                     SshMPIntegerConst op1,
                     SshMPIntegerConst op2)
{
  SSH_MP_WORKSPACE_DEFINE;
  SshWord *rem, *quot, *div;
  unsigned int rem_n, quot_n, bits;

  /* Check sizes first. */
  if (op1->n < op2->n)
    {
      ssh_mprz_set_ui(q, 0);
      ssh_mprz_set(r, op1);
      return;
    }

  if (op1->n == op2->n)
    {
      if (ssh_mpk_cmp(op1->v, op1->n, op2->v, op2->n) < 0)
        {
          ssh_mprz_set_ui(q, 0);
          ssh_mprz_set(r, op1);
          return;
        }
    }

  rem_n = op1->n + 1;
  quot_n = op1->n - op2->n + 1;

  /* Do some reallocation. */
  ssh_mprz_realloc(q, op1->n - op2->n + 1);
  ssh_mprz_realloc(r, op2->n);

  /* Allocate temporary space. */
  SSH_MP_WORKSPACE_ALLOC(rem, (rem_n + quot_n + op2->n));
  quot = rem + rem_n;
  div  = quot + quot_n;

  /* Clear and copy. */
  ssh_mpk_memzero(quot, quot_n);
  ssh_mpk_memcopy(rem, op1->v, op1->n);
  rem[op1->n] = 0;

  /* Normalize, this can be optimized. XXX */
  ssh_mpk_memcopy(div, op2->v, op2->n);

  bits = ssh_mpk_leading_zeros(div, op2->n);
  ssh_mpk_shift_up_bits(div, op2->n, bits);
  ssh_mpk_shift_up_bits(rem, rem_n, bits);

  /* Certify the length. */
  if (rem[rem_n - 1] == 0)
    rem_n--;

  /* Do the division iteration. */
  ssh_mpk_div(quot, quot_n, rem, rem_n, div, op2->n);

  /* Quotient is immediately correct. However, remainder must be
     denormalized. */
  ssh_mpk_shift_down_bits(rem, op2->n, bits);

  /* Now set the quotient. */
  ssh_mpk_memcopy(q->v, quot, quot_n);
  q->n = quot_n;

  /* Set the remainder. */
  ssh_mpk_memcopy(r->v, rem, op2->n);
  r->n = op2->n;

  /* Figure out quotient sign. */
  SSH_MP_XOR_SIGNS(q, op1, op2);

  /* Check sizes. */
  while (q->n && q->v[q->n - 1] == 0)
    q->n--;

  while (r->n && r->v[r->n - 1] == 0)
    r->n--;

  /* Handle the sign of the remainder. */
  if (SSH_MP_GET_SIGN(op1))
    SSH_MP_SET_SIGN(r);
  else
    SSH_MP_NO_SIGN(r);

  /* Make sure that zeros are positive :) */
  if (r->n == 0)
    SSH_MP_NO_SIGN(r);
  if (q->n == 0)
    SSH_MP_NO_SIGN(q);

  /* Free temporary storage. */
  SSH_MP_WORKSPACE_FREE(rem);
}

void ssh_mprz_div(SshMPInteger ret_q, SshMPIntegerConst op1,
                  SshMPIntegerConst op2)
{
  SshMPIntegerStruct t;
  ssh_mprz_init(&t);
  ssh_mprz_divrem(ret_q, &t, op1, op2);
  ssh_mprz_clear(&t);
}

/* Compute the remainder i.e. op1 (mod op2). */
void ssh_mprz_mod(SshMPInteger r, SshMPIntegerConst op1,
                SshMPIntegerConst op2)
{
  SSH_MP_WORKSPACE_DEFINE;
  SshWord *rem, *div;
  unsigned int rem_n, bits, div_n;

  /* Check sizes first. */
  if (op1->n == 0)
    {
      ssh_mprz_set_ui(r, 0);
      return;
    }

  if (op1->n < op2->n)
    {
      if (SSH_MP_GET_SIGN(op1))
        {
          ssh_mprz_add(r, op2, op1);
          return;
        }
      ssh_mprz_set(r, op1);
      return;
    }

  if (op1->n == op2->n)
    {
      if (ssh_mpk_cmp(op1->v, op1->n, op2->v, op2->n) < 0)
        {
          if (SSH_MP_GET_SIGN(op1))
            {
              ssh_mprz_add(r, op2, op1);
              return;
            }
          ssh_mprz_set(r, op1);
          return;
        }
    }

  rem_n = op1->n + 1;
  div_n = op2->n;

  /* Do some reallocation. */
  ssh_mprz_realloc(r, op2->n);

  /* Allocate temporary space. */
  SSH_MP_WORKSPACE_ALLOC(rem, (rem_n + div_n));
  div  = rem + rem_n;

  /* Clear and copy. */
  ssh_mpk_memcopy(rem, op1->v, op1->n);
  rem[op1->n] = 0;

  /* Normalize, this can be optimized. XXX */
  ssh_mpk_memcopy(div, op2->v, op2->n);

  bits = ssh_mpk_leading_zeros(div, op2->n);
  ssh_mpk_shift_up_bits(div, op2->n, bits);
  ssh_mpk_shift_up_bits(rem, rem_n, bits);

  /* Certify the length. */
  if (rem[rem_n - 1] == 0)
    rem_n--;

  /* Do the division iteration. */
  ssh_mpk_mod(rem, rem_n, div, op2->n);

  /* Quotient is immediately correct. However, remainder must be
     denormalized. */
  ssh_mpk_shift_down_bits(rem, op2->n, bits);

  /* Check sizes. */
  rem_n = op2->n;
  while (rem_n && rem[rem_n - 1] == 0)
    rem_n--;

  /* Handle possible negative input here. */
  if (SSH_MP_GET_SIGN(op1))
    {
      ssh_mpk_sub(rem, op2->v, op2->n, rem, rem_n);

      /* Check size again. */
      rem_n = op2->n;
      while (rem_n && rem[rem_n - 1] == 0)
        rem_n--;
    }

  /* Set the remainder. */
  r->n = rem_n;
  ssh_mpk_memcopy(r->v, rem, rem_n);
  SSH_MP_WORKSPACE_FREE(rem);

  /* Remainder has no sign (it is always positive). */
  SSH_MP_NO_SIGN(r);
}

/* Extra routines for special numbers. */

void ssh_mprz_mul_ui(SshMPInteger ret, SshMPIntegerConst op, SshWord u)
{
  SSH_MP_WORKSPACE_DEFINE;
  SshWord *temp;
  unsigned int temp_n;

  if (u == 0 || op->n == 0)
    {
      ssh_mprz_set_ui(ret, 0);
      return;
    }

  temp_n = op->n + 1;
  ssh_mprz_realloc(ret, temp_n);

  if (op->v != ret->v)
    temp = ret->v;
  else
    {
      SSH_MP_WORKSPACE_ALLOC(temp, temp_n);
    }

  ssh_mpk_memzero(temp, temp_n);

  /* Multiply. */
  ssh_mpk_mul_ui(temp, op->v, op->n, u);

  /* Finish the management. */
  if (temp != ret->v)
    {
      ssh_mpk_memcopy(ret->v, temp, temp_n);
      SSH_MP_WORKSPACE_FREE(temp);
    }

  ret->n = temp_n;

  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;

  SSH_MP_COPY_SIGN(ret, op);
}

SshWord ssh_mprz_divrem_ui(SshMPInteger q, SshMPIntegerConst op,
                           SshWord u)
{
  SSH_MP_WORKSPACE_DEFINE;
  SshWord *temp, *norm, t, rem;
  unsigned int temp_n, r;

  if (u == 0)
    ssh_fatal("ssh_mprz_divrem_ui: division by zero.");

  if (op->n == 0)
    {
      ssh_mprz_set_ui(q, 0);
      return 0;
    }

  /* Figure out the normalization of 'u'. */
  t = u;
  r = 0;
  SSH_MPK_COUNT_LEADING_ZEROS(r, t);
  t <<= r;

  /* Enlarge integers. */
  temp_n = op->n + 1;

  ssh_mprz_realloc(q, temp_n);

  if (q->v != op->v)
    temp = q->v;
  else
    {
      SSH_MP_WORKSPACE_ALLOC(temp, temp_n);
    }

  /* Normalize. */
  norm = ssh_xmalloc(sizeof(SshWord) * (op->n + 1));
  ssh_mpk_memcopy(norm, op->v, op->n);
  norm[op->n] = 0;

  ssh_mpk_shift_up_bits(norm, op->n + 1, r);

  rem = ssh_mpk_div_ui(temp, temp_n, norm, op->n + 1, t);

  ssh_xfree(norm);

  /* Correct remainder. */
  rem >>= r;

  /* Quotient is correct. */
  if (temp != q->v)
    {
      ssh_mpk_memcopy(q->v, temp, temp_n);
      SSH_MP_WORKSPACE_FREE(temp);
    }

  /* Set the size. */
  q->n = temp_n;

  while (q->n && q->v[q->n - 1] == 0)
    q->n--;

  if (q->n == 0)
    SSH_MP_NO_SIGN(q);

  return rem;
}

void ssh_mprz_div_ui(SshMPInteger q, SshMPIntegerConst op, SshWord u)
{
  ssh_mprz_divrem_ui(q, op, u);
}


SshWord ssh_mprz_mod_ui(SshMPIntegerConst op, SshWord u)
{
  SSH_MP_WORKSPACE_DEFINE;
  SshWord *norm, rem, t;
  unsigned int r;

  if (u == 0)
    ssh_fatal("ssh_mprz_div_ui: division by zero.");

  if (op->n == 0)
    return 0;

  /* Handle the normalization of 'u'. */
  t = u;
  r = 0;
  SSH_MPK_COUNT_LEADING_ZEROS(r, t);
  t <<= r;

  /* Allocate and normalize. */
  SSH_MP_WORKSPACE_ALLOC(norm, op->n + 1);
  ssh_mpk_memcopy(norm, op->v, op->n);
  norm[op->n] = 0;

  ssh_mpk_shift_up_bits(norm, op->n + 1, r);
  rem = ssh_mpk_mod_ui(norm, op->n + 1, t);

  SSH_MP_WORKSPACE_FREE(norm);

  /* Correct remainder. */
  rem >>= r;

  return rem;
}

/* GMP like interface to mod_ui. Just for compatibility. */
SshWord ssh_mprz_mod_ui2(SshMPInteger ret, SshMPIntegerConst op,
                         SshWord u)
{
  SshWord t;
  t = ssh_mprz_mod_ui(op, u);
  ssh_mprz_set_ui(ret, t);
  return t;
}

#if 0
/* Useful but messy dump function for integers. Can be used for
   debugging, but probably will be removed later. */
void ssh_mprz_dump(SshMPIntegerConst op)
{
  int i;
  printf("op: size = %u, allocated = %u, sign = %08x, words = \n",
         op->n, op->m, op->sign);
  printf("  ");
  for (i = op->n; i; i--)
    {
#if SIZEOF_LONG==4
      printf("%08lx ", op->v[i - 1]);
#else
      printf("%16lx ", op->v[i - 1]);
#endif /* SIZEOF_LONG==4 */
    }
  printf("\n");
  printf("  (0 ");
  for (i = op->n; i; i--)
    {
      printf("+ %lu*2^%u ", (unsigned long)op->v[i - 1],
             sizeof(unsigned long)*8*(i - 1));
    }
  printf(")\n");
#if 0
  {
    SshMPIntegerStruct temp;

    /* Use this only if you know that division by small integer works. */
    printf("backwards: ");
    ssh_mprz_init(&temp);
    ssh_mprz_set(&temp, op);
    while (ssh_mprz_cmp_ui(&temp, 0) != 0)
      {
        printf("%lu", ssh_mprz_div_ui(&temp, &temp, 10));
      }
    printf("\n");
  }
#endif
}
#else /* dump */
void ssh_mprz_dump(SshMPIntegerConst op)
{
  /* Do nothing. */
}
#endif

/* Miscellaneous, these will be useful later. */

/* Get a bit at position 'bit'. Returns thus either 1 or 0. */
unsigned int ssh_mprz_get_bit(SshMPIntegerConst op, unsigned int bit)
{
  unsigned int i;

  if (op->n == 0)
    return 0;

  /* Find out the amount of words. */
  i = bit / SSH_WORD_BITS;
  bit %= SSH_WORD_BITS;

  /* Too large. */
  if (i >= op->n)
    return 0;

  return (op->v[i] >> bit) & 0x1;
}

/* Set a bit at position 'bit'. */
void ssh_mprz_set_bit(SshMPInteger op, unsigned int bit)
{
  unsigned int i;

  /* Find out the amount of words. */
  i = bit / SSH_WORD_BITS;
  bit %= SSH_WORD_BITS;

  /* Allocate some new space and clear the extra space. */
  ssh_mprz_realloc(op, i + 1);
  ssh_mprz_clear_extra(op);

  op->v[i] |= ((SshWord)1 << bit);

  if (op->n < i + 1)
    op->n = i + 1;
}

/* Clear a bit at position 'bit'. */
void ssh_mp_clr_bit(SshMPInteger op, unsigned int bit)
{
  unsigned int i;
  /* Find out the word offset. */
  i = bit / SSH_WORD_BITS;
  if (i >= op->n)
    return;
  /* Find out the bit offset. */
  bit %= SSH_WORD_BITS;
  op->v[i] &= (~((SshWord)1 << bit));
  /* Normalize. */
  while (op->n > 0 && op->v[op->n-1] == 0)
    op->n--;
}

/* Compute the size of integer 'op' in base 'base'. Slow in many cases,
   but fast in base 2.  */
unsigned int ssh_mprz_get_size(SshMPIntegerConst op, SshWord base)
{
  unsigned int digits;
  SshMPIntegerStruct temp;

  switch (base)
    {
    case 0:
    case 1:
      return 0;
    case 2:
      /* Exact bit size quickly. */
      return ssh_mpk_size_in_bits(op->v, op->n);
    default:
      /* XXX Use division to divide to the base. Clearly this is slow, but
         this will be used only rarely. */
      ssh_mprz_init(&temp);
      ssh_mprz_set(&temp, op);
      if (ssh_mprz_cmp_ui(&temp, 0) < 0)
        ssh_mprz_neg(&temp, &temp);
      for (digits = 0; temp.n; digits++)
        ssh_mprz_div_ui(&temp, &temp, base);
      ssh_mprz_clear(&temp);
      return digits;
    }
}

/* Print routine. */

/* These are useful for hex and less bases. */
const unsigned char ssh_mprz_int_to_char[16] =
{ "0123456789abcdef" };

const unsigned char ssh_mprz_char_to_int[128] =
{
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
    0,   1,   2,   3,   4,   5,   6,   7,
    8,   9, 255, 255, 255, 255, 255, 255,
  255,  10,  11,  12,  13,  14,  15, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255,  10,  11,  12,  13,  14,  15, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
};

/* These are useful for bases upto hexes, that is most
   importantly base 64. */
const unsigned char ssh_mprz_int_to_base64[64] =
{ "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/" };

const unsigned char ssh_mprz_base64_to_int[128] =
{
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255,  62, 255, 255, 255,  63,
   52,  53,  54,  55,  56,  57,  58,  59,
   60,  61, 255, 255, 255, 255, 255, 255,
  255,   0,   1,   2,   3,   4,   5,   6,
    7,   8,   9,  10,  11,  12,  13,  14,
   15,  16,  17,  18,  19,  20,  21,  22,
   23,  24,  25, 255, 255, 255, 255, 255,
  255,  26,  27,  28,  29,  30,  31,  32,
   33,  34,  35,  36,  37,  38,  39,  40,
   41,  42,  43,  44,  45,  46,  47,  48,
   49,  50,  51, 255, 255, 255, 255, 255,
};

/* Transform the integer into a string format in base 'base'. */
char *ssh_mprz_get_str(SshMPIntegerConst op, SshWord base)
{
  SshMPIntegerStruct temp;
  unsigned int digits, real_digits, i, j, l;
  SshWord k, d;
  char *str;
  const unsigned char *table;
  Boolean sign = FALSE;

  /* Cannot handle larger than base 64 numbers nor smaller than 2. */
  if (base > 64 || base < 2)
    return NULL;

  if (base <= 16)
    table = ssh_mprz_int_to_char;
  else
    table = ssh_mprz_int_to_base64;

  if (ssh_mprz_cmp_ui(op, 0) == 0)
    {
      str = ssh_xmalloc(2);

      if (base <= 16)
        {
          str[0] = '0';
          str[1] = '\0';
        }
      else
        {
          str[0] = 'A';
          str[1] = '\0';
        }
      return str;
    }

  ssh_mprz_init(&temp);
  ssh_mprz_set(&temp, op);

  real_digits = digits = ssh_mprz_get_size(op, base);

  if (ssh_mprz_cmp_ui(&temp, 0) < 0)
    {
      digits++;
      sign = TRUE;
      ssh_mprz_neg(&temp, &temp);
    }

  switch (base)
    {
    case 8:
      digits++;
      break;
    case 16:
      digits += 2;
      break;
    case 64:
      digits++;
      break;
    default:
      break;
    }

  str = ssh_xmalloc(digits + 1);

  /* This is a very slow way to compute. We should atleast optimize this
     to take care of cases when base = 2^n. */

  for (j = 1, d = base; ;
       d = k, j++)
    {
      k = d * base;
      if (k / base != d)
        break;
    }

  for (i = 0; i < real_digits && temp.n; i += j)
    {
      k = ssh_mprz_divrem_ui(&temp, &temp, d);

      if (j + i > real_digits)
        j = real_digits - i;

      for (l = 0; l < j; l++)
        {
          str[(digits - (1 + i + l))] = table[k % base];
          k /= base;
        }
    }

  ssh_mprz_clear(&temp);

  /* Set the beginning to indicate the sign and base. */
  i = 0;
  if (sign)
    {
      str[0] = '-';
      i = 1;
    }

  switch (base)
    {
    case 8:
      str[i] = '0';
      break;
    case 16:
      str[i] = '0';
      str[i + 1] = 'x';
      break;
    case 64:
      str[i] = '#';
      break;
    default:
      break;
    }

  str[digits] = '\0';
  return str;
}

/* Convert a string into an integer in base 'base'. */
int ssh_mprz_set_str(SshMPInteger op, const char *str, SshWord base)
{
  size_t size = strlen(str);
  size_t i;
  const unsigned char *table;
  Boolean sign = FALSE;
  SshWord k, d, s;
  unsigned int j, l;

  /* Init with zero. */
  ssh_mprz_set_ui(op, 0);

  /* Skip leading whitespace and signs. */
  for (i = 0; i < size; i++)
    {
      switch (str[i])
        {
        case ' ':
        case '\t':
        case '\r':
        case '\n':
          break;
        case '-':
          if (!sign)
            {
              sign = TRUE;
              break;
            }
          return 0;
          break;
        case '0':
          /* Either base 8 or base 16. */
          if (tolower(str[i + 1]) == 'x')
            {
              /* Base-16. */
              if (base == 16 || base == 0)
                {
                  table = ssh_mprz_char_to_int;
                  base = 16;
                  i += 2;
                  goto read_number;
                }
            }
          if (isdigit((int)str[i + 1]))
            {
              /* Base-8 */
              if (base == 8 || base == 0)
                {
                  table = ssh_mprz_char_to_int;
                  base = 8;
                  i++;
                  goto read_number;
                }
            }
          if (base == 0)
            return 0;
          if (base <= 16)
            table = ssh_mprz_char_to_int;
          else
            table = ssh_mprz_base64_to_int;
          goto read_number;
          break;
        case '#':
          /* Base-64. */
          if (base == 64 || base == 0)
            {
              table = ssh_mprz_base64_to_int;
              base = 64;
              i++;
              goto read_number;
            }
          return 0;
          break;
        default:
          /* Any base or base-10 */
          if (base == 0)
            base = 10;
          if (base <= 16)
            table = ssh_mprz_char_to_int;
          else
            table = ssh_mprz_base64_to_int;
          goto read_number;
          break;
        }
    }

  /* No number to read. */
  return 0;

read_number:

  /* Generate large divisor. */
  for (j = 1, d = base;;
       d = k, j++)
    {
      k = d * base;
      if ((k / base) != d)
        break;
    }

  /* Loop through the string. */
  for (l = 0, k = 0; i <= size; i++)
    {
      switch (str[i])
        {
        case '\t':
        case ' ':
        case '\n':
          continue;
        }

      s = table[(unsigned char)(str[i] & 127)];
      if (s == 255)
        break;
      if (s >= base)
        break;

      k *= base;
      k += s;

      l++;
      if (l == j)
        {
          ssh_mprz_mul_ui(op, op, d);
          ssh_mprz_add_ui(op, op, k);
          l = 0;
          k = 0;
        }
    }

  /* Finish it off. */
  if (l)
    {
      for (i = 1, d = base; i < l; i++)
        d *= base;

      ssh_mprz_mul_ui(op, op, d);
      ssh_mprz_add_ui(op, op, k);
    }

  if (sign)
    ssh_mprz_neg(op, op);

  /* Return the number of limbs used. */
  return 1;
}

#if 1
/* Out integer to a stream. */
void ssh_mprz_out_str(FILE *fp, unsigned int base,
                      SshMPIntegerConst op)
{
  char *str;

  str = ssh_mprz_get_str(op, base);
  if (fp == NULL)
    fputs(str, stdout);
  else
    fputs(str, fp);
  ssh_xfree(str);
}
#endif

/* Faster routines of the above quick and dirty routines. */
void ssh_mprz_get_buf(unsigned char *buf, size_t buf_length,
                      SshMPIntegerConst op)
{
  size_t i, j, k;

  /* Set up k. */
  k = buf_length;

  /* Loop through all the words in the big number. */
  for (i = 0; i < op->n && k; i++)
    {
      SshWord w = op->v[i];

      /* Run through all the bytes of the input.

         Remark. In special cases (when k*8 >= SSH_WORD_BITS) we could
                 be slightly faster than this.
         */
      for (j = 0; j < SSH_WORD_BITS && k; j += 8, k--)
        {
          buf[k-1] = (unsigned char)(w & 0xff);
          w >>= 8;
        }
    }
  /* Clean up the rest of the buffer. */
  for (; k; k--)
    buf[k-1] = 0x0;
}

void ssh_mprz_set_buf(SshMPInteger ret, const unsigned char *buf,
                    size_t buf_length)
{
  size_t i, j, k;

  /* Compute the size of the buffer, in words. */
  k = (buf_length + (SSH_WORD_BITS/8) - 1)/(SSH_WORD_BITS/8);

  /* Reallocate enough space. */
  ssh_mprz_realloc(ret, k + 1);

  /* Set k as one beyond the buffer. */
  k = buf_length;

  for (i = 0; k; i++)
    {
      SshWord w;

      /* Build the word. */
      for (j = 0, w = 0; j < SSH_WORD_BITS && k; j += 8, k--)
        w += (((SshWord)buf[k-1]) << j);

      ret->v[i] = w;
    }

  ret->n = i;
  while (ret->n && ret->v[ret->n-1] == 0)
    ret->n--;
}

/* Random number routines. */

/* 'Find' a random number of 'bits' bits. */
void ssh_mprz_rand(SshMPInteger op, unsigned int bits)
{
  unsigned int i, k;

  /* Compute the word and bit positions. */
  k = bits / SSH_WORD_BITS;
  bits %= SSH_WORD_BITS;

  ssh_mprz_realloc(op, k + 1);

  /* Generate enough random bits. */
  for (i = 0; i < k + 1; i++)
    op->v[i] = ssh_rand();

  /* Don't do any shifting? */
  if (bits == 0)
    {
      op->n = k;
      while (op->n && op->v[op->n - 1] == 0)
        op->n--;
      SSH_MP_NO_SIGN(op);
      return;
    }

  /* Trivial shifting, masking... */
  op->v[k] = op->v[k] & (((SshWord)1 << bits) - 1);
  op->n = k + 1;

  while (op->n && op->v[op->n - 1] == 0)
    op->n--;
  SSH_MP_NO_SIGN(op);
}

/* Slow but allows one to build either very sparse (in 2-adic sense :) or
   very dense random numbers. */
void ssh_mprz_rand_w(SshMPInteger op, unsigned int bits, unsigned int weigth)
{
  unsigned int i, j;
  SshWord k, n0, n1;

  /* Clear the number before anything else. */
  ssh_mprz_set_ui(op, 0);
  ssh_mprz_clear_extra(op);

  /* We need some kind of an algorithm, which reasonably well builds
     sparse random numbers. They should be almost uniformly
     distributed. */

  /* One easy algorithm is to consider the probability that a bit is
     set, denoted by p = weigth/bits. However, we cannot work with
     floats so we have to do some tricks. */

  for (i = 0, j = 0; i < bits; i++)
    {
      /* Get a value from 0 <= k < 2^32. */
      k = ssh_rand();

      /* Convert k to comparable value. */

      SSH_MPK_LONG_MUL(n1, n0, k, (SshWord)bits);
      if (n1 <= (SshWord)weigth)
        ssh_mprz_set_bit(op, i);
    }
}

/* Slow, but so simple to write that I had to do it. */
void ssh_mprz_pow(SshMPInteger ret, SshMPIntegerConst g,
                SshMPIntegerConst e)
{
  SshMPIntegerStruct temp;
  unsigned int bits, i;

  /* Check the sign. */
  if (ssh_mprz_cmp_ui(e, 0) < 0)
    {
      /* This doesn't work! */
      ssh_fatal("ssh_mprf_pow:"
                " integers cannot be raised into a negative power.");
    }

  /* Trivial cases. */
  if (ssh_mprz_cmp_ui(e, 0) == 0)
    {
      ssh_mprz_set_ui(ret, 1);
      return;
    }

  if (ssh_mprz_cmp_ui(e, 1) == 0)
    {
      ssh_mprz_set(ret, g);
      return;
    }

  ssh_mprz_init(&temp);
  ssh_mprz_set(&temp, g);

  /* Compute the size of the exponent. */
  bits = ssh_mpk_size_in_bits(e->v, e->n);

  for (i = bits - 1; i; i--)
    {
      ssh_mprz_square(&temp, &temp);
      if (ssh_mprz_get_bit(e, i - 1))
        ssh_mprz_mul(&temp, &temp, g);
    }

  ssh_mprz_set(ret, &temp);
  ssh_mprz_clear(&temp);
}

void ssh_mprz_pow_ui_exp(SshMPInteger ret, SshMPIntegerConst g, SshWord e)
{
  SshMPIntegerStruct temp;

  /* Trivial cases. */
  if (e == 0)
    {
      ssh_mprz_set_ui(ret, 1);
      return;
    }

  if (e == 1)
    {
      ssh_mprz_set(ret, g);
      return;
    }

  ssh_mprz_init(&temp);
  ssh_mprz_set(&temp, g);
  ssh_mprz_set_ui(ret, 1);
  
  while (e)
    {
      if (e & 1)
        ssh_mprz_mul(ret, ret, &temp);
      e >>= 1; if (!e) break;
      ssh_mprz_square(&temp, &temp);
    }

  ssh_mprz_clear(&temp);
}

/* XXX Write optimized binary versions of the next two routines. That is,
   implement the binary gcd routines. They are reasonably simple, but
   I don't think they will be much faster thus it's easier to get
   along with these. */

/* Naive versions of these routines, which are fairly rarely used. */
void ssh_mprz_gcd(SshMPInteger d, SshMPIntegerConst a,
                  SshMPIntegerConst b)
{
  SshMPIntegerStruct a0, b0, r;

  ssh_mprz_init(&a0);
  ssh_mprz_init(&b0);
  ssh_mprz_init(&r);

  ssh_mprz_set(&a0, a);
  ssh_mprz_set(&b0, b);

  /* Standard gcd, however, we should implemented much faster ways also. */
  while (ssh_mprz_cmp_ui(&b0, 0) != 0)
    {
      ssh_mprz_mod(&r, &a0, &b0);
      ssh_mprz_set(&a0, &b0);
      ssh_mprz_set(&b0, &r);
    }

  ssh_mprz_set(d, &a0);

  ssh_mprz_clear(&a0);
  ssh_mprz_clear(&b0);
  ssh_mprz_clear(&r);
}

/* Compute (d, u, v) given (a, b) such that au + bv = d. */
void ssh_mprz_gcdext(SshMPInteger d, SshMPInteger u, SshMPInteger v,
                   SshMPIntegerConst a, SshMPIntegerConst b)
{
  SshMPIntegerStruct v1, v3, t1, t3, d0, u0, x;

  if (ssh_mprz_cmp_ui(b, 0) == 0)
    {
      ssh_mprz_set(d, a);
      ssh_mprz_set_ui(v, 0);
      ssh_mprz_set_ui(u, 1);
    }

  ssh_mprz_init(&v1);
  ssh_mprz_init(&v3);
  ssh_mprz_init(&t1);
  ssh_mprz_init(&t3);
  ssh_mprz_init(&u0);
  ssh_mprz_init(&d0);
  ssh_mprz_init(&x);

  ssh_mprz_set_ui(&u0, 1);
  ssh_mprz_set(&d0, a);
  ssh_mprz_set_ui(&v1, 0);
  ssh_mprz_set(&v3, b);

  /* Check for zero value using the internal size, which is bit ugly. */
  while (v3.n != 0)
    {
      /* Standard extended GCD algorithm inner loop. See for example
         Cohen's book. */
      ssh_mprz_divrem(&x, &t3, &d0, &v3);
      ssh_mprz_mul(&t1, &x, &v1);
      ssh_mprz_sub(&t1, &u0, &t1);
      ssh_mprz_set(&u0, &v1);
      ssh_mprz_set(&d0, &v3);
      ssh_mprz_set(&v1, &t1);
      ssh_mprz_set(&v3, &t3);
    }

  /* Compute v. */
  ssh_mprz_mul(&t1, a, &u0);
  ssh_mprz_sub(&t1, &d0, &t1);
  ssh_mprz_divrem(&v1, &v3, &t1, b);

  ssh_mprz_set(d, &d0);
  ssh_mprz_set(u, &u0);
  ssh_mprz_set(v, &v1);

  ssh_mprz_clear(&v1);
  ssh_mprz_clear(&v3);
  ssh_mprz_clear(&t1);
  ssh_mprz_clear(&t3);
  ssh_mprz_clear(&d0);
  ssh_mprz_clear(&u0);
  ssh_mprz_clear(&x);
}

/* Simple, but hopefully reasonably efficient. This is almost directly
   from Cohen's book. Improve if more speed is needed, one could open
   things up a bit, but this seems reasonably efficient. */
void ssh_mprz_sqrt(SshMPInteger sqrt_out, SshMPIntegerConst op)
{
  SshMPIntegerStruct x, y, r, t;
  int bits;

  /* Check impossible cases. */
  if (ssh_mprz_cmp_ui(op, 0) <= 0)
    {
      /* Should we terminate? Perhaps we return the integer part of this
         operation. */
      ssh_mprz_set_ui(sqrt_out, 0);
      return;
    }

  ssh_mprz_init(&x);
  ssh_mprz_init(&y);
  ssh_mprz_init(&r);
  ssh_mprz_init(&t);

  /* Find a nice estimate for n. */
  bits = ssh_mpk_size_in_bits(op->v, op->n);

  /* This should be fairly correct estimate. */
  ssh_mprz_set_bit(&x, (bits + 2)/2);

  /* Loop until a nice value found. */
  while (1)
    {
      /* Compute the newtonian step. */
      ssh_mprz_divrem(&t, &r, op, &x);
      ssh_mprz_add(&t, &t, &x);
      ssh_mprz_div_2exp(&y, &t, 1);

      if (ssh_mprz_cmp(&y, &x) < 0)
        ssh_mprz_set(&x, &y);
      else
        break;
    }

  /* Finished. */
  ssh_mprz_set(sqrt_out, &x);

  ssh_mprz_clear(&x);
  ssh_mprz_clear(&y);
  ssh_mprz_clear(&r);
  ssh_mprz_clear(&t);
}

/* Basic bit operations, for integers. These are simple, but useful
   sometimes. */
void ssh_mprz_and(SshMPInteger ret, SshMPIntegerConst op1,
                SshMPIntegerConst op2)
{
  unsigned int i;

  /* Swap. */
  if (op1->n > op2->n)
    {
      SshMPIntegerConst t;
      t = op1;
      op1 = op2;
      op2 = t;
    }

  /* Reallocate. */
  ssh_mprz_realloc(ret, op1->n);

  /* This can be written more optimally. */
  for (i = 0; i < op1->n; i++)
    ret->v[i] = op1->v[i] & op2->v[i];

  ret->n = op1->n;
  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;
  SSH_MP_NO_SIGN(ret);
}

void ssh_mprz_or(SshMPInteger ret, SshMPIntegerConst op1,
               SshMPIntegerConst op2)
{
  unsigned int i;

  /* Swap. */
  if (op1->n > op2->n)
    {
      SshMPIntegerConst t;
      t = op1;
      op1 = op2;
      op2 = t;
    }

  /* Reallocate. */
  ssh_mprz_realloc(ret, op2->n);

  /* This can be written more optimally. */
  for (i = 0; i < op1->n; i++)
    ret->v[i] = op1->v[i] | op2->v[i];
  for (; i < op2->n; i++)
    ret->v[i] = op2->v[i];

  ret->n = op2->n;
  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;
  SSH_MP_NO_SIGN(ret);
}

void ssh_mprz_xor(SshMPInteger ret, SshMPIntegerConst op1,
                SshMPIntegerConst op2)
{
  unsigned int i;

  /* Swap. */
  if (op1->n > op2->n)
    {
      SshMPIntegerConst t;
      t = op1;
      op1 = op2;
      op2 = t;
    }

  /* Reallocate. */
  ssh_mprz_realloc(ret, op1->n);

  /* This can be written more optimally. */
  for (i = 0; i < op1->n; i++)
    ret->v[i] = op1->v[i] ^ op2->v[i];
  for (; i < op2->n; i++)
    ret->v[i] = op2->v[i];

  ret->n = op2->n;
  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;
  SSH_MP_NO_SIGN(ret);
}

void ssh_mprz_not(SshMPInteger ret, SshMPIntegerConst op)
{
  unsigned int i;

  /* Reallocate. */
  ssh_mprz_realloc(ret, op->n);

  /* This can be written more optimally. */
  for (i = 0; i < op->n; i++)
    ret->v[i] = ~op->v[i];

  ret->n = op->n;
  while (ret->n && ret->v[ret->n - 1] == 0)
    ret->n--;
  SSH_MP_NO_SIGN(ret);
}

/* sshmp-integer.c */
